from c4v.scraper.crawler.crawlers.base_crawler import BaseCrawler
from c4v.scraper.crawler.crawlers.primicia_crawler import PrimiciaCrawler
from c4v.scraper.crawler.crawlers.el_pitazo_crawler import ElPitazoCrawler